Crate wrapper_lite

Crate wrapper_lite 

wrapper-lite

crates.io docs.rs tag

The newtype idiom gives compile time guarantees that the right type of value is supplied to a program.

This crate provides a declarative macro for you to build a wrapper type and implementing common traits for it.

For changes and migration guide, please refer to the CHANGELOG.

License

Licensed under either of:

at your option.

Macros

wrapper
Helper macro for building a wrapper type and implementing common traits for it.